/**
    Character Driver for Beckhoff BIOS API
    Author: Heiko Wilke <h.wilke@beckhoff.com>
    Author: Patrick Brünn <p.bruenn@beckhoff.com>
    Author: Chris Johns <chrisj@rtems.org>
    Copyright (C) 2013 - 2015  Beckhoff Automation GmbH

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <rtems.h>
#include <rtems/libio.h>
#include <rtems/bspIo.h>

#include "api.h"
#include "TcBaDevDef_bsd.h"

#define DRV_VERSION      "1.7"
#define DRV_DESCRIPTION  "Beckhoff BIOS API Driver"

/* Global Variables */
static struct bbapi_object g_bbapi =
{
	.mutex = PTHREAD_MUTEX_INITIALIZER
};
static rtems_device_major_number bbapi_major;

#if defined(__i386__)
static const uint64_t BBIOSAPI_SIGNATURE = 0x495041534F494242LL;	// API-String "BBIOSAPI"
#define BBIOSAPI_ATTR __attribute__ ((ms_abi, stdcall))
#elif defined(__x86_64__)
static const uint64_t BBIOSAPI_SIGNATURE = 0x3436584950414242LL;	// API-String "BBAPIX64"
#define BBIOSAPI_ATTR __attribute__ ((ms_abi))
#endif

/**
 * This function is a wrapper to the Beckhoff BIOS API entry function,
 * which uses MS Windows calling convention.
 * On x86_64 all we need to do is to use gcc "__attribute__((ms_abi))"
 */
typedef
    BBIOSAPI_ATTR uint32_t(*PFN_BBIOSAPI_CALL) (uint32_t group,
							   uint32_t offset,
							   void *in,
							   uint32_t inSize,
							   void *out,
							   uint32_t outSize,
							   uint32_t * bytes);

static unsigned int bbapi_call(void * const in,
			       void * const out,
			       PFN_BBIOSAPI_CALL entry,
			       const struct bbapi_struct *const cmd,
			       unsigned int *bytes_written)
{
	return entry(cmd->nIndexGroup, cmd->nIndexOffset, in,
		     cmd->nInBufferSize, out, cmd->nOutBufferSize,
		     (uint32_t*) bytes_written);
}

static uint32_t ioread32(void* addr)
{
  return *((volatile uint32_t*) addr);
}

static void mutex_lock(pthread_mutex_t* lock)
{
	int result = pthread_mutex_lock(lock);
	if (result != 0)
	  printk("BBAPI: lock error: %d (%d)\n", errno, result);
}

static void mutex_unlock(pthread_mutex_t* lock)
{
	(void) pthread_mutex_unlock(lock);
}

static unsigned int bbapi_rw(uint32_t group, uint32_t offset,
			     void * const in, uint32_t size_in,
			     void * const out, const uint32_t size_out)
{
	const struct bbapi_struct cmd = {
		.nIndexGroup = group,
		.nIndexOffset = offset,
		.pInBuffer = NULL,
		.nInBufferSize = size_in,
		.pOutBuffer = NULL,
		.nOutBufferSize = size_out
	};
	unsigned int bytes_written = 0;
	volatile unsigned int result = 0;
	mutex_lock(&g_bbapi.mutex);
	result = bbapi_call(in, out, g_bbapi.entry, &cmd, &bytes_written);
	mutex_unlock(&g_bbapi.mutex);
	return result;
}

unsigned int bbapi_read(uint32_t group, uint32_t offset,
			void * const out, const uint32_t size)
{
	return bbapi_rw(group, offset, NULL, 0, out, size);
}

unsigned int bbapi_write(uint32_t group, uint32_t offset,
			 void * const in, uint32_t size)
{
	return bbapi_rw(group, offset, in, size, NULL, 0);
}

int bbapi_board_is(const char *const boardname)
{
	char board[CXPWRSUPP_MAX_DISPLAY_LINE] = { 0 };

	bbapi_read(BIOSIGRP_GENERAL, BIOSIOFFS_GENERAL_GETBOARDNAME, &board,
		   sizeof(board) - 1);
	return 0 == strncmp(board, boardname, sizeof(board));
}

/**
 * bbapi_copy_bios() - Copy BIOS from SPI flash into RAM
 * @bbapi: pointer to a not initialized bbapi_object
 * @pos: pointer to the BIOS identifier string in flash
 *
 * We use BIOS shadowing to increase realtime performance.
 * The BIOS identifier string is followed by the 32-Bit BIOS API
 * function offset. This offset is the location of the BIOS API entry
 * function and is located at most 4096 bytes in front of the
 * BIOS memory end. So we calculate the size of the BIOS and copy it
 * from SPI Flash into RAM.
 * Accessing the BIOS in ROM while running realtime applications would
 * otherwise have bad effects on the realtime behaviour.
 *
 * Note: PAGE_KERNEL_EXEC omits the "no execute bit" exception
 *
 * Return: 0 for success, -ENOMEM if the allocation of kernel memory fails
 */
static int bbapi_copy_bios(struct bbapi_object *bbapi,
				  void * pos)
{
	const uint32_t offset = ioread32(pos + 8);
	const size_t size = offset + 4096;
	bbapi->memory = malloc(size);
	if (bbapi->memory == NULL) {
		printk("error: malloc for Beckhoff BIOS API failed\n");
		return -ENOMEM;
	}
	memcpy(bbapi->memory, pos, size);
	bbapi->entry = bbapi->memory + offset;
	return 0;
}

/**
 * bbapi_find_bios() - Find BIOS in SPI flash and copy it into RAM
 * @bbapi: pointer to a not initialized bbapi_object
 *
 * If successful bbapi->memory and bbapi->entry point to the bios in RAM
 *
 * Return: 0 if the bios was successfully copied into RAM
 */
static int bbapi_find_bios(struct bbapi_object *bbapi)
{
	static const size_t STEP_SIZE = 0x10;

	// Try to remap IO Memory to search the BIOS API in the memory
	void *const start = (void* const) BBIOSAPI_SIGNATURE_PHYS_START_ADDR;
	const void *const end = start + BBIOSAPI_SIGNATURE_SEARCH_AREA;
	int result = -EFAULT;
	void *pos;
	size_t off;

	if (start == NULL) {
		printk("wanring: Mapping memory search area for BIOS API failed\n");
		return -ENOMEM;
	}
	// Search through the remapped memory and look for the BIOS API String
	for (off = 0; off < STEP_SIZE; ++off) {
		for (pos = start + off; pos < end; pos += STEP_SIZE) {
			const uint32_t low = ioread32(pos);
			const uint32_t high = ioread32(pos + 4);
			const uint64_t lword = ((uint64_t) high << 32 | low);
			if (BBIOSAPI_SIGNATURE == lword) {
				result = bbapi_copy_bios(bbapi, pos);
				printk
				    ("BIOS found and copied from: %p + 0x%lx | %lu\n",
				     start, pos - start, off);
				goto cleanup;
			}
		}
	}
cleanup:
	return result;
}

/**
 * You have to hold the lock on bbapi->mutex when calling this function!!!
 */
static int bbapi_ioctl_mutexed(struct bbapi_object *const bbapi,
			       const struct bbapi_struct *const cmd)
{
	unsigned int written = 0;
	if (cmd->nInBufferSize > sizeof(bbapi->in)) {
		printk("error: %s(): nInBufferSize invalid\n", __FUNCTION__);
		return -EINVAL;
	}
	if (cmd->nOutBufferSize > sizeof(bbapi->out)) {
		printk("error: %s(): nOutBufferSize: %d invalid\n", __FUNCTION__,
		       (int) cmd->nOutBufferSize);
		return -EINVAL;
	}
	// BIOS can operate on kernel space buffers only -> make a temporary copy
	memcpy(bbapi->in, cmd->pInBuffer, cmd->nInBufferSize);
	// Call the BIOS API
	if (bbapi_call(bbapi->in, bbapi->out, bbapi->entry, cmd, &written)) {
		printk("error: %s(): call to BIOS failed\n", __FUNCTION__);
		return -EIO;
	}
	// Copy the BIOS output to the output buffer in user space
	memcpy(cmd->pOutBuffer, bbapi->out, written);
	return 0;
}

static rtems_device_driver
bbapi_ioctl(rtems_device_major_number major,
	    rtems_device_minor_number minor,
	    void                    * arg)
{
	rtems_libio_ioctl_args_t *args = arg;
	unsigned int cmd = args->command;
	struct bbapi_struct bbstruct;
	int result = -EINVAL;
	if (!g_bbapi.entry) {
		printk("%s(): not initialized.\n", __FUNCTION__);
		return -EINVAL;
	}
	// Check if IOCTL CMD matches BBAPI Driver Command
	if (cmd != BBAPI_CMD) {
		printk("Wrong Command\n");
		return -EINVAL;
	}
	// Copy data (BBAPI struct) from User Space to Kernel Module - if it fails, return error
	memcpy(&bbstruct, (const void *)arg, sizeof(bbstruct));

	if (bbstruct.nIndexOffset >= 0xB0) {
		printk("cmd: 0x%x : 0x%x not available from user mode\n",
		       (int) bbstruct.nIndexGroup, (int) bbstruct.nIndexOffset);
		return -EACCES;
	}

	mutex_lock(&g_bbapi.mutex);
	result = bbapi_ioctl_mutexed(&g_bbapi, &bbstruct);
	mutex_unlock(&g_bbapi.mutex);
	return result;
}

static rtems_device_driver
bbapi_open(rtems_device_major_number major,
	   rtems_device_minor_number minor,
	   void                    * arg)
{
	return RTEMS_SUCCESSFUL;
}

static rtems_device_driver
bbapi_close(rtems_device_major_number major RTEMS_UNUSED,
	    rtems_device_minor_number minor RTEMS_UNUSED,
	    void                    * arg)
{
	return RTEMS_SUCCESSFUL;
}

const rtems_driver_address_table bbapi_ops = {
  	.initialization_entry = NULL,
	.open_entry = bbapi_open,
	.close_entry = bbapi_close,
	.read_entry = NULL,
	.write_entry = NULL,
	.control_entry = bbapi_ioctl
};

static void update_display(void)
{
	char line[CXPWRSUPP_MAX_DISPLAY_LINE] = "RTEMS 7890123456";
	uint8_t enable = 0xff;
	int r;

	strncpy(line + 6, RTEMS_VERSION, sizeof(line) - 1 - 6);
	r = bbapi_write(BIOSIGRP_CXPWRSUPP,
			BIOSIOFFS_CXPWRSUPP_DISPLAYLINE2, line, sizeof(line));
	if (r != 0)
	  printk("d2 %04x\n", r);

	r = bbapi_read(BIOSIGRP_GENERAL, BIOSIOFFS_GENERAL_GETBOARDNAME, line,
		   sizeof(line));
	if (r != 0)
	  printk("gn %04x\n", r);

	r = bbapi_write(BIOSIGRP_CXPWRSUPP, BIOSIOFFS_CXPWRSUPP_DISPLAYLINE1, line,
		    sizeof(line));
	if (r != 0)
	  printk("d1 %04x\n", r);

	r = bbapi_write(BIOSIGRP_CXPWRSUPP,
		    BIOSIOFFS_CXPWRSUPP_ENABLEBACKLIGHT, &enable,
		    sizeof(enable));
	if (r != 0)
	  printk("bl %04x\n", r);
}

inline static bool bbapi_supports(uint32_t group, uint32_t offset)
{

	switch (bbapi_read(group, offset, NULL, 0)) {
	case BIOSAPI_INVALIDSIZE:
	case BIOSAPI_INVALIDPARM:
		return true;
	default:
		return false;
	}
}

#define bbapi_supports_display() \
	bbapi_supports(BIOSIGRP_CXPWRSUPP, BIOSIOFFS_CXPWRSUPP_ENABLEBACKLIGHT)

#define bbapi_supports_power() \
	bbapi_supports(BIOSIGRP_CXPWRSUPP, BIOSIOFFS_CXPWRSUPP_GETTYPE)

#define bbapi_supports_sups() \
	(bbapi_supports(BIOSIGRP_SUPS, BIOSIOFFS_SUPS_GPIO_PIN_EX) \
	 || bbapi_supports(BIOSIGRP_SUPS, BIOSIOFFS_SUPS_GPIO_PIN))

int bbapi_init(void)
{
	int               result;
	rtems_status_code sc;

	printf("%s, %s\n", DRV_DESCRIPTION, DRV_VERSION);

	result = bbapi_find_bios(&g_bbapi);
	if (result) {
		printk("Beckhoff BIOS API not available on this System\n");
		return result;
	}
	sc = rtems_io_register_driver(0, &bbapi_ops, &bbapi_major);
	if (sc != RTEMS_SUCCESSFUL) {
		goto rollback_memory;
	}

	if (bbapi_supports_display()) {
		update_display();
	}
	return 0;

rollback_memory:
	free(g_bbapi.memory);
	return result;
}
